/* 
   Author: Joel W. Reed (joelreed@yahoo.com).  Copyright 1999.
   This file is distributed  under the Artistic License.
   See the file LICENSE for details.
*/

#include <stdio.h>
#include <math.h>
#include <string>
#include "argparse.h"
#include "nn-structs.h"

#define ERRMSG(msg) if (!silent) fprintf(stderr, msg);
const char TAG_STDIN[] = "standard input";
const char TAG_STDOUT[] = "standard output";
int silent=0, verbose=0;

fpnum_t activation(fpnum_t h)
{
	//const static double steepness = 0.5;
	// the below should really be 1/(1+exp(-2*steepness*h))
	// but since i'm setting steepness to .5 (which works well for 0/1 range)
	// i've simplified the expression below for speed
	if (h<-30) return 0.0; // prevent overflow TODO: set this to an accurate CUTOFF
	return 1.0/(1.0+exp(-h));
}

fpnum_t compute_delta(fpnum_t obtained, fpnum_t delta)
{
	// again steepness set to .5 (see note in function activation())
	// which simplifies this equation
	return ((obtained * (1 - obtained)) + 0.1) * delta;
}

unit::unit(const char* t, fpnum_t threshold, fpnum_t v) : _value(v), 
	_threshold (threshold), _delta(0.0)
{
	_tag = t ? t : unique_name();
}

const char* unit::unique_name()
{
	static unsigned u=0;
	static char buf[256]; 
	
	buf[0]='\0';
	sprintf(buf, "<hidden-%d>", u);
	
	++u;
	return buf;
}

connection::connection(fpnum_t w, unit* f, unit* s)  : _weight(w), _first(f), _second(s)
{
}

network::network()
{
}

bool network::initialize(const std::vector<std::string>& l1_tags, const std::vector<std::string>& ln_tags, const std::vector<int>& dims)
{
	if (verbose) printf("initializing the network...\n");

	// create unit layer 1 from tag list
	{
		unit_layer l;
		for (std::vector<std::string>::const_iterator i = l1_tags.begin(); i != l1_tags.end(); ++i)
		{
			l.push_back(unit(i->c_str()));
			if (verbose) printf("  added layer one node %s...\n", i->c_str());
		}       
		_unit_layers.push_back(l);
	}

	// create unit layers 2->N based on requested dimensions
	for (std::vector<int>::const_iterator i = dims.begin(); i != dims.end(); ++i)
	{
		unit_layer l;
		for (int j=0; j < *i; ++j) l.push_back(unit());
		_unit_layers.push_back(l);

		if (verbose) printf("  added hidden layer with %d nodes...\n", *i);
	}

	// create unit layer 'N' from tag list
	{
		unit_layer l;
		for (std::vector<std::string>::const_iterator i = ln_tags.begin(); i != ln_tags.end(); ++i)
		{
			l.push_back(unit(i->c_str()));
			if (verbose) printf("  added layer N node %s...\n", i->c_str());
		}       
		_unit_layers.push_back(l);
	}

	// now connect the layers with the weights
	for (unit_layers_t::iterator i = _unit_layers.begin(), j = i+1; j != _unit_layers.end(); ++i, ++j)
	{
		weight_layer wl;

		for (unit_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			double multiplier = 1/sqrt(i->size());
			for (unit_layer::iterator m = j->begin(); m != j->end(); ++m)
			{
				double wv = (rand()/((double)RAND_MAX))*multiplier;
				connection c(wv, &(*k), &(*m));
				wl.push_back(c);
			}
		}

		_weight_layers.push_back(wl);
	}

	if (verbose) printf("done.\n\n");
}

bool network::load(const char* fn)
{
	if (verbose) printf("loading %s...\n", (fn)? fn : TAG_STDIN);

	// pattern can be read from stdin, not a saved network. 
	FILE* file = fopen(fn, "r");

	while (!feof(file))
	{
		char type[256]; 
		if (0==fgets(type, sizeof(type), file))
			break;

		if (strstr(type, "unit "))
		{
			unit_layer layer;
			int cnt = atoi(type+5); // get to cnt. yes i know this is ugly
			
			while (!feof(file) && cnt > 0)
			{
				char tag[MAX_TAG_SIZE];
				fpnum_t threshold; fscanf(file, "%s %lf\n", tag, &threshold);
				layer.push_back(unit(tag, threshold));

				if (verbose) printf("  adding node %s %lf to network.\n", tag, threshold);
				--cnt;
			}

			_unit_layers.push_back(layer);
		}
		else if (strstr(type, "weight "))
		{
			weight_layer layer;
			int cnt = atoi(type+7);
			
			while (!feof(file) && cnt > 0)
			{
				char num[256]; fgets(num, sizeof(num), file);
				fpnum_t value = atof(num);
				layer.push_back(connection(value));

				if (verbose) printf("  adding weight %lf\n", value);
				--cnt;
			}
			
			_weight_layers.push_back(layer);
		}
	}

	// now connect the layers with the weights
	weight_layers_t::iterator wli = _weight_layers.begin();
	for (unit_layers_t::iterator i = _unit_layers.begin(), j = i+1; j != _unit_layers.end(); ++i, ++j, ++wli)
	{
		weight_layer::iterator wi = wli->begin();
		for (unit_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			for (unit_layer::iterator m = j->begin(); m != j->end(); ++m, ++wi)
			{
				wi->_first = &(*k);
				wi->_second = &(*m);
			}
		}
	}

	fclose(file);
	if (verbose) printf("done.\n\n");
	return true;
}

bool network::save(const char* fn) const
{
	if (verbose) printf("saving %s...\n", (fn)? fn : TAG_STDOUT);

	FILE* file = ((0==fn)||('-' == *fn))? stdout : fopen(fn, "w");

	// save the unit values
	int cnt = 1; const int indx_of_last = _unit_layers.size();
	for (unit_layers_t::const_iterator i = _unit_layers.begin(); i != _unit_layers.end(); ++i, ++cnt)
	{
		fprintf(file, "unit %d\n", i->size());
		for (unit_layer::const_iterator k = i->begin(); k != i->end(); ++k)
		{
			fprintf(file, "%s %lf\n", k->_tag.c_str(), k->_threshold);
		}
	}

	// save the weights
	cnt = 0;
	for (weight_layers_t::const_iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i, ++cnt)
	{
		fprintf(file, "weight %d\n", i->size());
		for (weight_layer::const_iterator k = i->begin(); k != i->end(); ++k)
		{
			fprintf(file, "%lf\n", k->_weight);
		}
	}

	if ((0!=fn)&&('-' != *fn)) fclose(file);
	if (verbose) printf("done.\n\n");
	return true;
}

bool network::process(const unit_layer& inputs)
{
	if (verbose) printf("processing %d inputs...\n", inputs.size());

	// get reference to input layer of nn
	unit_layer& t = *(_unit_layers.begin()); 

	// apply input pattern to first layer of network...
	for (unit_layer::const_iterator i = inputs.begin(); i != inputs.end(); ++i)
	{
 		bool unused = true;
		for (unit_layer::iterator j = t.begin(); j != t.end(); ++j)
		{
 			if (j->_tag == i->_tag)
			{
				if (verbose) printf("  node %s receiving input of %lf\n", i->_tag.c_str(), i->_value);
				j->_value = i->_value; 
				unused = false;
				break;
			}
		}       
		
		if ((unused)&&(verbose)) fprintf(stderr, "warning: network input layer does not have an unit for '%s'\n", i->_tag.c_str()); 
	}

	// propogate the signal thru the network...
	for (weight_layers_t::iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i)
	{
		// first sum the weights
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			k->_second->_value += (k->_first->_value * k->_weight);  
			if (verbose) printf("  adding unit=%lf * weight=%lf to node %s -> %lf\n",  k->_first->_value, k->_weight, k->_second->_tag.c_str(), k->_second->_value);
		}
	}

	for (unit_layers_t::iterator i = _unit_layers.begin()+1; i < _unit_layers.end(); ++i)
	{
		for (unit_layer::iterator j = i->begin(); j != i->end(); ++j)
		{
			// then apply the activation function
			j->_value += j->_threshold;
			if (verbose) printf("  sum=%lf, threshold=%lf, activation=%lf\n", j->_value, j->_threshold, activation(j->_value));
			j->_value = activation(j->_value);
		}
	}

	if (verbose) printf("done.\n\n");
	return true;
}

bool network::learn(const char* target)
{
	if (verbose) printf("outputs...\n");

	// get reference to output layer of nn
	unit_layer& t = *(_unit_layers.rbegin()); 
	
	for (unit_layer::iterator i = t.begin(); i != t.end(); ++i)
	{
		fpnum_t desired = (i->_tag == target)? 1 : 0;
		fpnum_t diff = desired - i->_value;
		if (!silent) printf ("  %lf\n", diff);
		i->_delta = compute_delta(i->_value, diff);
	}
	
	if (verbose) printf("done.\n\n");
	
	// propogate the signal thru the network...
	if (verbose) printf("learning...\n");

	for (weight_layers_t::reverse_iterator i = _weight_layers.rbegin(); i != _weight_layers.rend(); ++i)
	{
		if (verbose) printf("\n  processing weight layer with size %d\n", i->size());

		// first sum the deltas
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			if (verbose) printf("  2nd-delta=%lf  * weight=%lf + 1st-delta=%lf = ", 
			    k->_second->_delta, k->_weight, k->_first->_delta);
			k->_first->_delta += (k->_second->_delta * k->_weight);  
			if (verbose) printf("  %lf for 1st node=%s\n", k->_first->_delta, k->_first->_tag.c_str()); 
		}

		// then apply the delta function
		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			if (verbose) printf("  1st-value=%lf, 1st-delta=%lf, ",
					    k->_first->_value, k->_first->_delta);
			k->_first->_delta = compute_delta(k->_first->_value, k->_first->_delta);
			if (verbose) printf("  computed-delta=%lf\n", k->_first->_delta); 
		}
	}

	// now adjust our weights based on the deltas
	if (verbose) printf("done.\n\n");
	if (verbose) printf("adjusting weights...\n");

	unit_layers_t::const_iterator ui = _unit_layers.begin();
	for (weight_layers_t::iterator i = _weight_layers.begin(); i != _weight_layers.end(); ++i, ++ui)
	{
		if (verbose) printf("  processing weight layer with size %d\n", i->size());

		for (weight_layer::iterator k = i->begin(); k != i->end(); ++k)
		{
			//TODO: change 1/unit_cnt to some other adaptive pattern (page 124..)
			if (verbose) printf("  %lf %lf , fan-in %d\n", k->_first->_value, k->_first->_delta, ui->size());
			
			fpnum_t bias_delta = k->_first->_delta * (1.0/ui->size());
			k->_second->_threshold += bias_delta;

			fpnum_t weight_delta =  k->_first->_value * bias_delta;
			k->_weight += weight_delta;
		}
	}

	if (verbose) printf("done.\n\n");
	return true;
}

bool pattern::load(const char* fn)
{
	if (verbose) printf("loading pattern from %s...\n", (fn)? fn : TAG_STDIN);

	// load input pattern...
	FILE* file = ((0==fn)||(*fn=='-'))? stdin: fopen(fn, "r");

	while (!feof(file))
	{
		char tag[MAX_TAG_SIZE];
		fpnum_t value; fscanf(file, "%s %lf ", tag, &value);
		if (verbose) printf("  pattern::load -> %s %lf\n", tag, value);
		push_back(unit(tag, 0, value));
	}

	if (size()==0)
	{
		ERRMSG("Error: No input data found.\n");
		return false;
	}

	if ((0!=fn)&&('-' != *fn)) fclose(file);

	if (verbose) printf("done.\n\n");
	return true;
}

int main (int argc, char* argv[])
{
	if (argc<2) 
	{
		printf("%s version %s \nCopyright (c) 1999 by Joel W. Reed. \nTry '%s --help' or '%s --usage' for more information.\n", PACKAGE, VERSION, PACKAGE, PACKAGE);
		exit(0);
	}

	arguments args;
	argp_parse (&parser, argc, argv, 0, 0, &args);
	silent = args.silent;
	verbose = args.verbose;

	// seed random number generator
	srand(time(0));

	// for now if a network is being created from scratch we don't also do a process/learn. could change later
	network nn;
	if (args.nn_dims.size()>0)
	{
		nn.initialize(args.l1_tags, args.ln_tags, args.nn_dims);
		nn.save(args.out_fname);
		exit(0);
	}
	else
	{
		if (!nn.load(args.network_fname)) 
		{
			ERRMSG("Error: Failed to load neural network. Please confirm the file is appropriately formatted.\n");
			return 1;
		}

		pattern learning_pattern; 
		if (!learning_pattern.load(args.pattern_fname))
		{
			ERRMSG("Error: Failed to load learning pattern. Please confirm the file is appropriately formatted.\n");
			return 1;
		}

		nn.process(learning_pattern); // compute output
		nn.learn(args.target); // adjust weights based on deviance

		nn.save(args.out_fname);
	}
	
	return 0;
}
